import numpy as np
import clust.scripts.datastructures as ds
import sklearn.metrics.pairwise as skdists
import clust.scripts.numeric as nu
import clust.scripts.statistical as st
import sys
from clust.scripts.mnplots import mnplotsdistancethreshold


def reorderClusters(B, X, GDM, returnOrderIndices = False):
    Bloc = np.array(B)
    Xloc = ds.listofarrays2arrayofarrays(X)

    Bloc = Bloc[:, np.any(Bloc, axis=0)]  # Only keep non-empty clusters

    B_ordered = np.zeros(Bloc.shape, dtype=bool)
    K = Bloc.shape[1]  # Number of clusters
    L = Xloc.shape[0]  # Number of datasets

    if K == 0:
        return Bloc

    # Find Cmeans and distances between clusters
    Cmeans = np.array([None] * L, dtype=object)
    D = np.full([K, K, L], np.inf)  # KxKxL  (initialised with inf values)
    for l in range(L):
        Cmeans[l] = np.zeros([K, Xloc[l].shape[1]], dtype=float)  # (K) x (X[l] samples)
        for k in range(K):
            Cmeans[l][k] = np.mean(Xloc[l][Bloc[GDM[:, l], k], :], axis=0)

        # For empty clusters, the distances are inf (as filled in the initialisation of D),
        # For non-empty clusters, the distances are calculated by skdists.euclidean_distances between cmeans
        I_not_empty_c = ~np.array([np.any(np.isnan(cm)) for cm in Cmeans[l]])  # Indices of clusters with nans (empty)
        I_not_empty_c = np.where(I_not_empty_c)[0]  # From boolean indices to integer indices
        non_empty_c_dists = skdists.euclidean_distances(Cmeans[l][I_not_empty_c])  # K_not_empty x K_not_empty
        for nec in range(len(I_not_empty_c)):
            D[I_not_empty_c[nec], I_not_empty_c, l] = non_empty_c_dists[nec]
        #D[:, :, l] = skdists.euclidean_distances(Cmeans[l])  # KxK
    D = np.median(D, axis=2)  # KxK

    # Set first cluster as first, then find closest by closest
    B_ordered[:, 0] = Bloc[:, 0]
    I = np.zeros(K, dtype=int)
    I[0] = 0
    clustersDone = np.zeros(K, dtype=bool)
    clustersDone[0] = True
    for k in range(1,K):
        relevantD = D[I[k-1], ~clustersDone]
        clustersLeft = np.nonzero(~clustersDone)[0]
        nextCluster = np.argmin(relevantD)
        nextCluster = clustersLeft[nextCluster]
        B_ordered[:, k] = Bloc[:, nextCluster]
        I[k] = nextCluster
        clustersDone[nextCluster] = True

    if returnOrderIndices:
        return (B_ordered, I)
    else:
        return B_ordered


def correcterrors_withinworse(B, X, GDM, falsepositivestrimmed=0.01):
    Bloc = np.array(B)
    Xloc = ds.listofarrays2arrayofarrays(X)

    [Ng, K] = Bloc.shape  # Ng genes and K clusters
    L = Xloc.shape[0]  # L datasets

    # Find clusters' means (Cmeans), absolute shifter clusters genes (SCG),
    # and the emperical CDF functions for them (cdfs)
    Cmeans = np.array([None] * L, dtype=object)
    SCG = np.array([None] * L, dtype=object)
    for l in range(L):
        Cmeans[l] = np.zeros([K, Xloc[l].shape[1]])  # K clusters x D dimensions
        SCG[l] = np.zeros([np.sum(np.sum(Bloc[GDM[:, l], :], axis=0)), Xloc[l].shape[1]])  # M* genes x D dimensions ...
        # (M* are all # genes in any cluster)

        gi = 0
        for k in range(K):
            Cmeans[l][k] = np.median(Xloc[l][Bloc[GDM[:, l], k], :], axis=0)
            csize = np.sum(Bloc[GDM[:, l], k])
            tmpSCG = nu.subtractaxis(Xloc[l][Bloc[GDM[:, l], k], :], Cmeans[l][k], axis=0)
            SCG[l][gi:(gi+csize),:] = np.abs(tmpSCG)
            gi += csize
        SCG[l] = SCG[l][np.any(SCG[l], axis=1)]  # Remove all zeros genes (rows of SCG[l])
        SCG[l] = np.sort(SCG[l], axis=0)
        if falsepositivestrimmed > 0:
            trimmed = int(falsepositivestrimmed * SCG[l].shape[0])
            if trimmed > 0:
                SCG[l] = SCG[l][0:-trimmed]  # trim the lowest (trimmed) rows in SCG

    # Helping function
    def iswithinworse(ref, x):
        return x <= np.max(ref)

    # Find who belongs
    belongs = np.ones([Ng, K, L], dtype=bool)  # Ng genes x K clusters x L datasets
    for l in range(L):
        for k in range(K):
            for d in range(Xloc[l].shape[1]):
                tmpX = np.abs(Xloc[l][:, d] - Cmeans[l][k, d])
                belongs[GDM[:, l], k, l] &= iswithinworse(SCG[l][:, d], tmpX)

    # Include in clusters genes which belongs everywhere
    B_out = np.all(belongs, axis=2)

    # Genes included in two clusters, include them in the closest in terms of its worst distance to any of the clusters
    # (guarrantee that the worst belongingness of a gene to a cluster is optimised)
    f = np.nonzero(np.sum(B_out, axis=1) > 1)[0]
    for fi in f:
        ficlusts = np.nonzero(B_out[fi])[0]  # Clusters competing over gene fi
        fidatasets = np.nonzero(GDM[fi])[0]  # Datasets that have gene fi
        localdists = np.zeros([len(ficlusts), len(fidatasets)])  # (Clusts competing) x (datasets that have fi)
        for l in range(len(fidatasets)):
            ll = fidatasets[l]  # Actual dataset index
            fi_ll = np.sum(GDM[:fi, ll])  # Index of fi in this Xloc[ll]
            localdists[:, l] = nu.dist_matrices(Cmeans[ll][ficlusts], Xloc[ll][fi_ll]).reshape([len(ficlusts)])
        localdists = np.max(localdists, axis=1)  # (Clusts competing) x 1
        ficlosest = np.argmin(localdists)  # Closest cluster
        B_out[fi] = False
        B_out[fi, ficlusts[ficlosest]] = True

    return B_out


def correcterrors_weighted_fixed_fpr(B, X, GDM, clustdists=None, falsepositivestrimmed=0.01, smallestClusterSize=11):
    Bloc = np.array(B)
    Xloc = ds.listofarrays2arrayofarrays(X)

    [Ng, K] = Bloc.shape  # Ng genes and K clusters
    L = Xloc.shape[0]  # L datasets

    # Normalise clustdists to provide weights. If not provided, make it unity for all
    if clustdists is None:
        clustdistsnorm = np.ones(K)
    else:
        clustweights = np.min(clustdists) / clustdists

    # Find clusters' means (Cmeans), absolute shifted clusters genes (SCG),
    # and the emperical CDF functions for them (cdfs)
    Cmeans = np.array([None] * L, dtype=object)
    SCG = np.array([None] * L, dtype=object)
    for l in range(L):
        Cmeans[l] = np.zeros([K, Xloc[l].shape[1]])  # K clusters x D dimensions
        SCG[l] = np.zeros([np.sum(np.sum(Bloc[GDM[:, l], :], axis=0)), Xloc[l].shape[1]])  # M* genes x D dimensions ...
        # (M* are all # genes in any cluster)

        gi = 0
        for k in range(K):
            Cmeans[l][k] = np.median(Xloc[l][Bloc[GDM[:, l], k], :], axis=0)
            csize = np.sum(Bloc[GDM[:, l], k])
            tmpSCG = nu.subtractaxis(Xloc[l][Bloc[GDM[:, l], k], :], Cmeans[l][k], axis=0)
            # MAIN ADDITION (SCG is weighted by clustweights, max clustweight is unity, and it is the best)
            tmpSCG = tmpSCG * clustweights[k]
            SCG[l][gi:(gi + csize), :] = np.abs(tmpSCG)
            gi += csize
        SCG[l] = SCG[l][np.any(SCG[l], axis=1)]  # Remove all zeros genes (rows of SCG[l])
        SCG[l] = np.sort(SCG[l], axis=0)
        if falsepositivestrimmed > 0:
            trimmed = int(falsepositivestrimmed * SCG[l].shape[0])
            if trimmed > 0:
                SCG[l] = SCG[l][0:-trimmed]  # trim the lowest (trimmed) rows in SCG

    # Helping function
    def iswithinworse(ref, x):
        return x <= np.max(ref)

    # Find who belongs
    belongs = np.ones([Ng, K, L], dtype=bool)  # Ng genes x K clusters x L datasets
    for l in range(L):
        for k in range(K):
            for d in range(Xloc[l].shape[1]):
                tmpX = np.abs(Xloc[l][:, d] - Cmeans[l][k, d])
                belongs[GDM[:, l], k, l] &= iswithinworse(SCG[l][:, d], tmpX)

    # Include in clusters genes which belongs everywhere
    B_out = np.all(belongs, axis=2)

    # Solve genes included in two clusters:
    solution = 2
    if solution == 1:
        # Genes included in two clusters, include them in the closest in terms of its worst distance to any of the clusters
        # (guarrantee that the worst belongingness of a gene to a cluster is optimised)
        f = np.nonzero(np.sum(B_out, axis=1) > 1)[0]
        for fi in f:
            ficlusts = np.nonzero(B_out[fi])[0]  # Clusters competing over gene fi
            fidatasets = np.nonzero(GDM[fi])[0]  # Datasets that have gene fi
            localdists = np.zeros([len(ficlusts), len(fidatasets)])  # (Clusts competing) x (datasets that have fi)
            for l in range(len(fidatasets)):
                ll = fidatasets[l]  # Actual dataset index
                fi_ll = np.sum(GDM[:fi, ll])  # Index of fi in this Xloc[ll]
                localdists[:, l] = nu.dist_matrices(Cmeans[ll][ficlusts], Xloc[ll][fi_ll]).reshape([len(ficlusts)])
            localdists = np.max(localdists, axis=1)  # (Clusts competing) x 1
            ficlosest = np.argmin(localdists)  # Closest cluster
            B_out[fi] = False
            B_out[fi, ficlusts[ficlosest]] = True
    elif solution == 2:
        # Genes included in two clusters, include them in the earlier cluster (smallest k)
        f = np.nonzero(np.sum(B_out, axis=1) > 1)[0]
        for fi in f:
            ficlusts = np.nonzero(B_out[fi])[0]  # Clusters competing over gene fi
            ficlosest = np.argmin(ficlusts)  # earliest cluster (smallest k)
            B_out[fi] = False
            B_out[fi, ficlusts[ficlosest]] = True



    # Remove clusters smaller than minimum cluster size
    ClusterSizes = np.sum(B_out, axis=0)
    B_out = B_out[:, ClusterSizes >= smallestClusterSize]

    return B_out


def correcterrors_weighted_outliers(B, X, GDM, clustdists=None, stds=3, smallestClusterSize=11):
    Bloc = np.array(B)
    Xloc = ds.listofarrays2arrayofarrays(X)

    [Ng, K] = Bloc.shape  # Ng genes and K clusters
    L = Xloc.shape[0]  # L datasets

    # Normalise clustdists to provide weights. If not provided, make it unity for all
    if clustdists is None:
        clustweights = np.ones(K)
    else:
        clustweights = np.min(clustdists) / clustdists

    # Find clusters' means (Cmeans), absolute shifted clusters genes (SCG),
    # and the emperical CDF functions for them (cdfs)
    Cmeans = np.array([None] * L, dtype=object)
    SCG = np.array([None] * L, dtype=object)
    for l in range(L):
        Cmeans[l] = np.zeros([K, Xloc[l].shape[1]])  # K clusters x D dimensions
        SCG[l] = np.zeros([np.sum(np.sum(Bloc[GDM[:, l], :], axis=0)), Xloc[l].shape[1]])  # M* genes x D dimensions ...
        # (M* are all # genes in any cluster)

        gi = 0
        for k in range(K):
            Cmeans[l][k] = np.median(Xloc[l][Bloc[GDM[:, l], k], :], axis=0)
            csize = np.sum(Bloc[GDM[:, l], k])
            tmpSCG = nu.subtractaxis(Xloc[l][Bloc[GDM[:, l], k], :], Cmeans[l][k], axis=0)
            # MAIN ADDITION (SCG is weighted by clustweights, max clustweight is unity, and it is the best)
            tmpSCG = tmpSCG * clustweights[k]
            SCG[l][gi:(gi + csize), :] = np.abs(tmpSCG)
            gi += csize
        SCG[l] = SCG[l][np.any(SCG[l], axis=1)]  # Remove all zeros genes (rows of SCG[l])
        SCG[l] = np.sort(SCG[l], axis=0)
        SCGmeans = np.mean(SCG[l], axis=0)
        SCGstds = np.std(SCG[l], axis=0)
        SCGouts = nu.divideaxis(nu.subtractaxis(SCG[l], SCGmeans, axis=0), SCGstds, axis=0)  # No. of stds away
        SCGouts = SCGouts > stds  # TRUE for outliers and FALSE for others (bool: M* genex x D dimensions)
        SCG[l][SCGouts] = 0.0  # Set the outlier values to zeros so they do not affect decisions later on

    # Helping function
    def iswithinworse(ref, x):
        return x <= np.max(ref)

    # Find who belongs
    belongs = np.ones([Ng, K, L], dtype=bool)  # Ng genes x K clusters x L datasets
    for l in range(L):
        for k in range(K):
            for d in range(Xloc[l].shape[1]):
                tmpX = np.abs(Xloc[l][:, d] - Cmeans[l][k, d])
                belongs[GDM[:, l], k, l] &= iswithinworse(SCG[l][:, d], tmpX)

    # Include in clusters genes which belongs everywhere
    B_out = np.all(belongs, axis=2)

    # Solve genes included in two clusters:
    solution = 2
    if solution == 1:
        # Genes included in two clusters, include them in the closest in terms of its worst distance to any of the clusters
        # (guarrantee that the worst belongingness of a gene to a cluster is optimised)
        f = np.nonzero(np.sum(B_out, axis=1) > 1)[0]
        for fi in f:
            ficlusts = np.nonzero(B_out[fi])[0]  # Clusters competing over gene fi
            fidatasets = np.nonzero(GDM[fi])[0]  # Datasets that have gene fi
            localdists = np.zeros([len(ficlusts), len(fidatasets)])  # (Clusts competing) x (datasets that have fi)
            for l in range(len(fidatasets)):
                ll = fidatasets[l]  # Actual dataset index
                fi_ll = np.sum(GDM[:fi, ll])  # Index of fi in this Xloc[ll]
                localdists[:, l] = nu.dist_matrices(Cmeans[ll][ficlusts], Xloc[ll][fi_ll]).reshape([len(ficlusts)])
            localdists = np.max(localdists, axis=1)  # (Clusts competing) x 1
            ficlosest = np.argmin(localdists)  # Closest cluster
            B_out[fi] = False
            B_out[fi, ficlusts[ficlosest]] = True
    elif solution == 2:
        # Genes included in two clusters, include them in the earlier cluster (smallest k)
        f = np.nonzero(np.sum(B_out, axis=1) > 1)[0]
        for fi in f:
            ficlusts = np.nonzero(B_out[fi])[0]  # Clusters competing over gene fi
            ficlosest = np.argmin(ficlusts)  # earliest cluster (smallest k)
            B_out[fi] = False
            B_out[fi, ficlusts[ficlosest]] = True



    # Remove clusters smaller than minimum cluster size
    ClusterSizes = np.sum(B_out, axis=0)
    B_out = B_out[:, ClusterSizes >= smallestClusterSize]

    return B_out


'''This version uses weighted mean and std of SCG rather than giving weights to the errors themselves'''
def correcterrors_weighted_outliers2(B, X, GDM, clustdists=None, stds=3, smallestClusterSize=11):
    Bloc = np.array(B)
    Xloc = ds.listofarrays2arrayofarrays(X)

    [Ng, K] = Bloc.shape  # Ng genes and K clusters
    L = Xloc.shape[0]  # L datasets

    # Normalise clustdists to provide weights. If not provided, make it unity for all
    if clustdists is None:
        clustweights = np.ones(K)
    else:
        clustweights = np.min(clustdists) / clustdists

    # Find clusters' means (Cmeans), absolute shifted clusters genes (SCG),
    # and the emperical CDF functions for them (cdfs)
    Cmeans = np.array([None] * L, dtype=object)
    SCG = np.array([None] * L, dtype=object)
    for l in range(L):
        Cmeans[l] = np.zeros([K, Xloc[l].shape[1]])  # K clusters x D dimensions
        SCG[l] = np.zeros([np.sum(np.sum(Bloc[GDM[:, l], :], axis=0)), Xloc[l].shape[1]])  # M* genes x D dimensions ...
        w = np.zeros(np.sum(np.sum(Bloc[GDM[:, l], :], axis=0)))  # M* genes
        # (M* are all # genes in any cluster)

        gi = 0
        for k in range(K):
            Cmeans[l][k] = np.median(Xloc[l][Bloc[GDM[:, l], k], :], axis=0)
            csize = np.sum(Bloc[GDM[:, l], k])
            tmpSCG = nu.subtractaxis(Xloc[l][Bloc[GDM[:, l], k], :], Cmeans[l][k], axis=0)
            SCG[l][gi:(gi + csize), :] = np.abs(tmpSCG)
            # Added this in this version
            w[gi:(gi + csize)] = clustweights[k]
            gi += csize
        SCG[l] = SCG[l][np.any(SCG[l], axis=1)]  # Remove all zeros genes (rows of SCG[l])
        SCG[l] = np.sort(SCG[l], axis=0)
        SCGmeans = np.average(SCG[l], weights=w, axis=0)
        SCGstds = st.weighted_std_axis(SCG[l], weights=w, axix=0)
        SCGouts = nu.divideaxis(nu.subtractaxis(SCG[l], SCGmeans, axis=0), SCGstds, axis=0)  # No. of stds away
        SCGouts = SCGouts > stds  # TRUE for outliers and FALSE for others (bool: M* genex x D dimensions)
        SCG[l][SCGouts] = 0.0  # Set the outlier values to zeros so they do not affect decisions later on

    # Helping function
    def iswithinworse(ref, x):
        return x <= np.max(ref)

    # Find who belongs
    belongs = np.ones([Ng, K, L], dtype=bool)  # Ng genes x K clusters x L datasets
    for l in range(L):
        for k in range(K):
            for d in range(Xloc[l].shape[1]):
                tmpX = np.abs(Xloc[l][:, d] - Cmeans[l][k, d])
                belongs[GDM[:, l], k, l] &= iswithinworse(SCG[l][:, d], tmpX)

    # Include in clusters genes which belongs everywhere
    B_out = np.all(belongs, axis=2)

    # Solve genes included in two clusters:
    solution = 2
    if solution == 1:
        # Genes included in two clusters, include them in the closest in terms of its worst distance to any of the clusters
        # (guarrantee that the worst belongingness of a gene to a cluster is optimised)
        f = np.nonzero(np.sum(B_out, axis=1) > 1)[0]
        for fi in f:
            ficlusts = np.nonzero(B_out[fi])[0]  # Clusters competing over gene fi
            fidatasets = np.nonzero(GDM[fi])[0]  # Datasets that have gene fi
            localdists = np.zeros([len(ficlusts), len(fidatasets)])  # (Clusts competing) x (datasets that have fi)
            for l in range(len(fidatasets)):
                ll = fidatasets[l]  # Actual dataset index
                fi_ll = np.sum(GDM[:fi, ll])  # Index of fi in this Xloc[ll]
                localdists[:, l] = nu.dist_matrices(Cmeans[ll][ficlusts], Xloc[ll][fi_ll]).reshape([len(ficlusts)])
            localdists = np.max(localdists, axis=1)  # (Clusts competing) x 1
            ficlosest = np.argmin(localdists)  # Closest cluster
            B_out[fi] = False
            B_out[fi, ficlusts[ficlosest]] = True
    elif solution == 2:
        # Genes included in two clusters, include them in the earlier cluster (smallest k)
        f = np.nonzero(np.sum(B_out, axis=1) > 1)[0]
        for fi in f:
            ficlusts = np.nonzero(B_out[fi])[0]  # Clusters competing over gene fi
            ficlosest = np.argmin(ficlusts)  # earliest cluster (smallest k)
            B_out[fi] = False
            B_out[fi, ficlusts[ficlosest]] = True



    # Remove clusters smaller than minimum cluster size
    ClusterSizes = np.sum(B_out, axis=0)
    B_out = B_out[:, ClusterSizes >= smallestClusterSize]

    return B_out


def optimise_tukey_sqrtSCG(B, X, GDM, clustdists=None, smallestClusterSize=11, tails=1, Q3s=2):
    Bloc = np.array(B)
    Xloc = ds.listofarrays2arrayofarrays(X)

    [Ng, K] = Bloc.shape  # Ng genes and K clusters
    L = Xloc.shape[0]  # L datasets

    # Normalise clustdists to provide weights. If not provided, make it unity for all
    if clustdists is None:
        clustdistsloc = np.ones(K)
    else:
        clustdistsloc = [c for c in clustdists]

    # Find clusters' means (Cmeans), absolute shifted clusters genes (SCG),
    # and the emperical CDF functions for them (cdfs)
    Cmeans = np.array([None] * L, dtype=object)
    SCG = np.array([None] * L, dtype=object)

    Cgood = mnplotsdistancethreshold(clustdistsloc, method='largestgap')
    for l in range(L):
        Cmeans[l] = np.zeros([K, Xloc[l].shape[1]])  # K clusters x D dimensions
        SCG[l] = np.zeros([np.sum(np.sum(Bloc[GDM[:, l], :], axis=0)), Xloc[l].shape[1]])  # M* genes x D dimensions ...
        w = np.zeros(np.sum(np.sum(Bloc[GDM[:, l], :], axis=0)))  # M* genes
        # (M* are all # genes in any cluster)

        gi = 0
        for k in range(K):
            if k in Cgood:
                Cmeans[l][k] = np.median(Xloc[l][Bloc[GDM[:, l], k], :], axis=0)
                csize = np.sum(Bloc[GDM[:, l], k])
                tmpSCG = nu.subtractaxis(Xloc[l][Bloc[GDM[:, l], k], :], Cmeans[l][k], axis=0)
                SCG[l][gi:(gi + csize), :] = np.abs(tmpSCG)
                gi += csize
            else:
                Cmeans[l][k] = np.empty(Xloc[l].shape[1])
                Cmeans[l][k][:] = np.nan
        SCG[l] = SCG[l][np.any(SCG[l], axis=1)]  # Remove all zeros genes (rows of SCG[l])

        if ds.numel(SCG[l] > 0):
            if tails == 1:
                Q3 = np.percentile(SCG[l], q=75, axis=0)
                thresh = Q3s * Q3
                SCGouts = SCG[l] > np.array([thresh for ii in range(0, SCG[l].shape[0])])
                SCG[l][SCGouts] = 0.0  # Set the outlier values to zeros so they do not affect decisions later on
            elif tails == 2:
                Q1 = np.percentile(np.sqrt(SCG[l]), q=25, axis=0)
                Q3 = np.percentile(np.sqrt(SCG[l]), q=75, axis=0)
                IQR = np.subtract(Q3, Q1)
                thresh = np.add(Q3, 1.5 * IQR)
                SCGouts = np.sqrt(SCG[l]) > np.array([thresh for ii in range(0, SCG[l].shape[0])])
                SCG[l][SCGouts] = 0.0  # Set the outlier values to zeros so they do not affect decisions later on
            else:
                raise ValueError('Invalid number of tails. It should be either 1 or 2.')
        else:
            SCG[l] = np.zeros((1, SCG[l].shape[1]))

    # Clusters mins and maxes (NEW)
    Cmins = np.array([None] * L, dtype=object)
    Cmaxes = np.array([None] * L, dtype=object)
    for l in range(L):
        Cmins[l] = np.zeros([K, Xloc[l].shape[1]])  # K clusters x D dimensions
        Cmaxes[l] = np.zeros([K, Xloc[l].shape[1]])  # K clusters x D dimensions
        for k in range(K):
            Cmins[l][k] = Cmeans[l][k] - np.max(SCG[l], axis=0)
            Cmaxes[l][k] = Cmeans[l][k] + np.max(SCG[l], axis=0)

    # Resolve overlaps between clusters (NEW)
    for k1 in range(K):
        for k2 in range(K):
            # Compare the pair of clusters only once, and don't compare a cluster with itself. This if statement
            # guarantees that k2 will always be a later cluster than k1.
            if (k1 >= k2):
                continue
            # Value of the smallest overlap between the ranges of the clusters k1 and k2, and ...
            # the dataset (l) and the dimension (d), at which this overlap is found
            # t_smallest overlap is the type of the overlap, (-1, 0, 1, or 2). Type (-1) means that the entire (min
            # to max) range of one cluster is within the range of the other cluster. This is the worse overlap.
            # Type (0) means that the max of (k1) is within the range of (min to max) of (k2), and type (1) is the other
            # way around. Type (2) means there is no overlap. This is the best and finding one of it breaks the loop
            v_smallestoverlap = 0
            l_smallestoverlap = -1
            d_smallestoverlap = -1
            t_smallestoverlap = -1  # Overlap type, read above
            for l in range(L):
                Nd = len(Cmins[l][k1])  # Dimensions in this dataset
                for d in range(Nd):
                    x1 = Cmaxes[l][k1][d]
                    x2 = Cmaxes[l][k2][d]
                    n1 = Cmins[l][k1][d]
                    n2 = Cmins[l][k2][d]
                    if (x1 > n2 and x1 <= x2):
                        if (n1 < n2):
                            ov = x1 - n2
                            if (t_smallestoverlap == -1 or ov < v_smallestoverlap):
                                t_smallestoverlap = 0
                                v_smallestoverlap = ov
                                l_smallestoverlap = l
                                d_smallestoverlap = d
                    elif (x2 > n1 and x2 <= x1):
                        if (n2 < n1):
                            ov = x2 - n1
                            if (t_smallestoverlap == -1 or ov < v_smallestoverlap):
                                t_smallestoverlap = 1
                                v_smallestoverlap = ov
                                l_smallestoverlap = l
                                d_smallestoverlap = d
                    else:
                        t_smallestoverlap = 2
                        continue  # Absolutely no overlap at this point, so k1 and k2 are distinct, so continue
                if (t_smallestoverlap == 2):
                    continue  # Absolutely no overlap at some point, so k1 and k2 are distinct, so continue

            # Sort out the overlap if exists between k1 and k2
            if (t_smallestoverlap == -1):
                # Here one of the two clusters always swallows the other one. So effectively remove the later one (k2).
                # Cluster removal is by making its minimum larger than its maximum at a single point (at l=0, d=0),
                # so effectively no gene will ever be mapped to it!
                Cmins[0][k2][0] = 1
                Cmaxes[0][k2][0] = 0
            elif (t_smallestoverlap == 0):
                Cmins[l_smallestoverlap][k2][d_smallestoverlap] = \
                    Cmaxes[l_smallestoverlap][k1][d_smallestoverlap] + sys.float_info.epsilon
            elif (t_smallestoverlap == 1):
                Cmaxes[l_smallestoverlap][k2][d_smallestoverlap] = \
                    Cmins[l_smallestoverlap][k1][d_smallestoverlap] - sys.float_info.epsilon

    # Find who belongs (NEW)
    belongs = np.ones([Ng, K, L], dtype=bool)  # Ng genes x K clusters x L datasets
    for l in range(L):
        for k in range(K):
            tmp1 = nu.largerthanaxis(Xloc[l], Cmins[l][k], axis=0, orequal=True)
            tmp2 = nu.lessthanaxis(Xloc[l], Cmaxes[l][k], axis=0, orequal=True)
            belongs[GDM[:, l], k, l] = np.all(np.logical_and(tmp1, tmp2), axis=1)


    # # Helping function (OLD - to be removed)
    # def iswithinworse(ref, x):
    #     return x <= np.max(ref)
    #
    # # Find who belongs (OLD - to be removed)
    # belongs = np.ones([Ng, K, L], dtype=bool)  # Ng genes x K clusters x L datasets
    # for l in range(L):
    #     for k in range(K):
    #         for d in range(Xloc[l].shape[1]):
    #             tmpX = np.abs(Xloc[l][:, d] - Cmeans[l][k, d])
    #             belongs[GDM[:, l], k, l] &= iswithinworse(SCG[l][:, d], tmpX)

    # Include in clusters genes which belongs everywhere (OLD - to be removed)
    B_out = np.all(belongs, axis=2)

    # Solve genes included in two clusters (OLD - should not be needed now - TO BE REMOVED)
    solution = 2
    if solution == 1:
        # Genes included in two clusters, include them in the closest in terms of its worst distance to any of the clusters
        # (guarrantee that the worst belongingness of a gene to a cluster is optimised)
        f = np.nonzero(np.sum(B_out, axis=1) > 1)[0]
        for fi in f:
            ficlusts = np.nonzero(B_out[fi])[0]  # Clusters competing over gene fi
            fidatasets = np.nonzero(GDM[fi])[0]  # Datasets that have gene fi
            localdists = np.zeros([len(ficlusts), len(fidatasets)])  # (Clusts competing) x (datasets that have fi)
            for l in range(len(fidatasets)):
                ll = fidatasets[l]  # Actual dataset index
                fi_ll = np.sum(GDM[:fi, ll])  # Index of fi in this Xloc[ll]
                localdists[:, l] = nu.dist_matrices(Cmeans[ll][ficlusts], Xloc[ll][fi_ll]).reshape([len(ficlusts)])
            localdists = np.max(localdists, axis=1)  # (Clusts competing) x 1
            ficlosest = np.argmin(localdists)  # Closest cluster
            B_out[fi] = False
            B_out[fi, ficlusts[ficlosest]] = True
    elif solution == 2:
        # Genes included in two clusters, include them in the earlier cluster (smallest k)
        f = np.nonzero(np.sum(B_out, axis=1) > 1)[0]
        for fi in f:
            ficlusts = np.nonzero(B_out[fi])[0]  # Clusters competing over gene fi
            ficlosest = np.argmin(ficlusts)  # earliest cluster (smallest k)
            B_out[fi] = False
            B_out[fi, ficlusts[ficlosest]] = True



    # Remove clusters smaller than minimum cluster size
    ClusterSizes = np.sum(B_out, axis=0)
    B_out = B_out[:, ClusterSizes >= smallestClusterSize]

    return B_out


import collections
import numpy as np
from copy import deepcopy


def length(x):
    if isinstance(x, (list, tuple, np.ndarray)):
        return len(x)
    elif x is None:
        return 0
    else:
        return 1


def numel(x):
    y = np.array(x)
    try:
        size = 1
        for dim in np.shape(y):
            size *= dim
    except:
        size = 1
    return size


def mat2vec(X, dim=1):
    if dim == 1:
        return np.reshape(X, [numel(X), 1])
    elif dim == 2:
        return np.reshape(X, [1, numel(X)])
    else:
        raise ValueError('dim should be either (1) for a column vector or (2) for a row vector')


def matlablike_index2D(X, I0, I1):
    """
    :param X: 2D array or 2D numpy.ndarray
    :param I0: row indices (integers or boolean)
    :param I1: column indices (integers or boolean)
    :return: numpy.ndarray of indexed values
    """
    Y = np.array(X)
    if isinstance(I0, str) and I0 in ['all', 'a']:
        I0 = [n for n in range(Y.shape[0])]
    if isinstance(I1, str) and I1 in ['all', 'a']:
        I1 = [n for n in range(Y.shape[1])]
    if numel(I0) > 1 and all(isinstance(n, bool) for n in I0):
        Y = Y[np.nonzero(I0)[0]]
    else:
        Y = Y[I0]
    if numel(I1) > 1 and all(isinstance(n, bool) for n in I1):
        Y = Y[:, np.nonzero(I1)[0]]
    else:
        Y = Y[:, I1]
    return Y


def matlablike_indexND(X, indices):
    Y = np.array(X)
    for i in range(len(indices)):
        if type(indices[i]).__module__ == np.__name__ and indices[i].dtype.type == np.bool_:
            Y = Y.take(np.nonzero(indices[i])[0], i)
        elif numel(indices[i]) > 1 and all(isinstance(n,bool) for n in indices[i]):
            Y = Y.take(np.nonzero(indices[i])[0], i)
        else:
            Y = Y.take(indices[i], i)
    return Y


def resolveargumentpairs(refparamsnames, defaults, params=()):
    res = ()
    p = params[0::2]
    v = params[1::2]
    for refp in refparamsnames:
        if refp in p:
            res = res.__add__((v[p.index(refp)],))
        else:
            res = res.__add__((defaults[refparamsnames.index(refp)],))
    return res


def listofarrays2arrayofarrays(X):
    # Check if 2D
    is2d = True
    N2 = -1
    for x in X:
        if isinstance(x, list):
            if N2 == -1:
                N2 = len(x)
            elif len(x) != N2:
                    is2d = False
                    break
        else:
            is2d = False
            break

    #Calculate
    L = len(X)
    if is2d:
        Xout = np.zeros([L, N2], dtype=object)
        for i in range(L):
            for j in range(N2):
                Xout[i,j] = deepcopy(X[i][j])
    else:
        Xout = np.zeros(L, dtype=object)
        for i in range(L):
            Xout[i] = deepcopy(X[i])

    #Return
    return Xout


def maxDepthOfArray(X):
    Xloc = deepcopy(X)
    if numel(Xloc) == 0:
        return 0
    if isinstance(Xloc, np.ndarray) and Xloc.dtype not in [object, np.ndarray, list, tuple]:
        return len(Xloc.shape)
    elif isinstance(Xloc, (list, tuple, np.ndarray)):
        return max([maxDepthOfArray(x) for x in Xloc]) + 1
    else:
        return 0


def reduceToArrayOfNDArraysAsObjects(X, N):
    Xloc = listofarrays2arrayofarrays(X)
    Xnew = np.array([], dtype=object)
    if maxDepthOfArray(Xloc) == N + 1:
        Xnew = deepcopy(Xloc)
    elif maxDepthOfArray(Xloc) < N + 1:
        Xnew = deepcopy(Xloc)
        for i in range(N + 1 - maxDepthOfArray(Xloc)):
            Xnew = np.expand_dims(Xnew, axis=0)
    else:
        for x in Xloc:
            N0 = maxDepthOfArray(x)
            if N0 == N:
                Xnew = np.concatenate((Xnew, listofarrays2arrayofarrays([Xloc])))
            elif N0 < N:
                Xtmp = x
                for i in range(N - N0):
                    Xtmp = np.expand_dims(Xtmp, axis=0)
                Xnew = np.concatenate((Xnew, Xtmp))
            else:
                Xnew = np.concatenate((Xnew, reduceToArrayOfNDArraysAsObjects(x, N)))
    return Xnew


def flattenAList(l):
    return [item for sublist in l for item in sublist]


def concatenateStrings(listofstrings, delim=', '):
    if isinstance(listofstrings, str):
        return listofstrings + ''
    elif len(listofstrings) == 1:
        return listofstrings[0] + ''
    else:
        res = listofstrings[0]
        for s in listofstrings[1:]:
            res += delim + s
        return res


def findArrayInAnotherArray1D(x, y):
    """
    :param x:
    :param y:
    :return: Array with a length similar to x showing x's elements' first occurance indices in y.
    -1 is given for unfound elements.
    """
    index = np.argsort(y)
    sorted_y = y[index]
    sorted_index = np.searchsorted(sorted_y, x)

    xindex = np.take(index, sorted_index, mode="clip")
    mask = y[xindex] != x

    return np.array(np.ma.array(xindex, mask=mask).tolist(fill_value=-1))


def findArrayInSubArraysOfAnotherArray1D(x, y):
    x = np.array(x)
    y = np.array(y)
    y_unfolded = np.array([], dtype=y.dtype)
    y_indices_of_y_unfolded = np.array([], dtype=int)
    for yi in range(len(y)):
        y_unfolded = np.append(y_unfolded, y[yi], axis=0)
        y_indices_of_y_unfolded = np.append(y_indices_of_y_unfolded, [yi for i in range(len(y[yi]))], axis=0)

    I = findArrayInAnotherArray1D(x, y_unfolded)
    I[I > -1] = y_indices_of_y_unfolded[I[I > -1]]
    return I





import numpy as np
import datetime as dt
import collections as collec
import clust.scripts.datastructures as ds
from clust.scripts.glob import outputwidth, version
import pandas as pd



def msgformated(msg, alignment='<', withnewline=True):
    # alignment: '<' left, '>' right, '^' centred, and '*^' centred while padding with ******
    msgsplit = msg.split('\n')
    res = ''
    for l in msgsplit:
        while len(l) > outputwidth:
            # Cut the line at the last space or tab before the max length (outputwidth); otherwise cut at outputwith
            indexoflastspace = outputwidth - l[outputwidth-1::-1].find(' ') - 1
            indexoflasttab = outputwidth - l[outputwidth - 1::-1].find('\t') - 1
            if indexoflastspace == outputwidth and indexoflasttab == outputwidth:
                indexofcut = outputwidth  # This is the index at which cut happens
                excludeindexofcut = False  # Don't remove the character at this index as it is not a space or a tab
            elif indexoflastspace < outputwidth and indexoflasttab < outputwidth:
                indexofcut = max(indexoflastspace, indexoflasttab)
                excludeindexofcut = True
            else:
                indexofcut = min(indexoflastspace, indexoflasttab)
                excludeindexofcut = True
            res += '| {{0:{0}{1}}} |\n'.format(alignment, outputwidth).format(l[:indexofcut])
            if excludeindexofcut:
                l = l[indexofcut+1:]
            else:
                l = l[indexofcut:]
        # Add what remains of the line
        res += '| {{0:{0}{1}}} |\n'.format(alignment, outputwidth).format(l)

    if not withnewline:
        res = res[:-1]

    return res


def arraytostring(A, delim=' ', openbrac='', closebrac=''):
    L = len(A)

    if L == 0:
        res = '{0}{1}'.format(openbrac, closebrac)
    else:
        res = '{0}{1}'.format(openbrac, A[0])
        for i in range(1,L):
            res = '{0}{1}{2}'.format(res, delim, A[i])
        res = '{0}{1}'.format(res, closebrac)

    return res


def topline(withnewline=True):
    res = '/{{0:=^{0}}}\\'.format(outputwidth+2).format('')
    if withnewline:
        res += '\n'
    return res


def bottomline(withnewline=True):
    res = '\\{{0:=^{0}}}/'.format(outputwidth+2).format('')
    if withnewline:
        res += '\n'
    return res


def midline(withnewline=True):
    res = '+{{0:-^{0}}}+'.format(outputwidth+2).format('')
    if withnewline:
        res += '\n'
    return res


def generateinitialmessage():
    starttime = dt.datetime.now()

    res = '\n' + topline()
    tmptxt = 'Clust\n' \
             '(Optimised consensus clustering of multiple heterogenous datasets)\n' \
             'Python package version {0} (2022) Basel Abu-Jamous'.format(version[1:])
    res += msgformated(tmptxt, alignment='^')
    res += midline()

    tmptxt = 'Analysis started at: {0}'.format(starttime.strftime('%A %d %B %Y (%H:%M:%S)'))
    res += msgformated(tmptxt, withnewline=False)
    return res, starttime


def generateoutputsummaryparag(X, Xprocessed, Map, GDMall, GDM, uncle_res, mn_res, B_corrected, starttime):
    # Sort out the ending time and consumed time
    endtime = dt.datetime.now()
    t = endtime - starttime
    d = t.days
    h = int(t.seconds / 3600)
    m = int((t.seconds % 3600) / 60)
    s = t.seconds % 60
    if t.total_seconds() < 1:
        mil = t.microseconds / 1000
        timeConsumedTxt = '{0} millisecond{1}'.format(mil, '' if mil == 1 else 's')
    else:
        timeConsumedTxt = '{0}{1}{2} {3}, {4} {5}, and {6} {7}' \
                          ''.format(d if d > 0 else '', ' days, ' if d > 1 else ' day, ' if d == 1 else '',
                                    h, 'hour' if h == 1 else 'hours',
                                    m, 'minute' if m == 1 else 'minutes',
                                    s, 'second' if s == 1 else 'seconds')

    # Generate the text
    L = len(X)
    label0 = 'genes' if Map is None else 'OrthoGroups (OGs)*'
    label1 = 'gene' if Map is None else 'OG'
    dataset_datasets = 'dataset' if L == 1 else 'datasets'
    Ng_original = GDMall.shape[0]
    Ng_to_clustering = GDM.shape[0]
    Ng_in_clusters = np.sum(np.any(B_corrected, axis=1))

    tmptxt = 'Analysis finished at: {0}\n' \
             'Total time consumed: {1}\n'.format(endtime.strftime('%A %d %B %Y (%H:%M:%S)'), timeConsumedTxt)
    res = midline()
    res += msgformated(tmptxt)
    res += bottomline()
    res += '\n' + topline()
    res += msgformated('RESULTS SUMMARY', alignment='^')
    res += midline()

    K = B_corrected.shape[1]

    if K > 0:
        tmptxt = 'Clust received {2} {3} with {4} unique {0}. After filtering, {5} {1}s made it to the clustering ' \
                 'step. Clust generated {6} clusters of {1}s, which in total include {7} {1}s. The smallest cluster ' \
                 'includes {8} {1}s, the largest cluster includes {9} {1}s, and the average cluster size ' \
                 'is {10} {1}s.' \
                 ''.format(label0, label1, L, dataset_datasets, Ng_original, Ng_to_clustering, K, Ng_in_clusters,
                           np.min(np.sum(B_corrected, axis=0)), np.max(np.sum(B_corrected, axis=0)),
                           int(round(np.mean(np.sum(B_corrected, axis=0)))))
    else:
        tmptxt = 'Clust received {2} {3} with {4} unique {0}. After filtering, {5} {1}s made it to the clustering ' \
                 'step. Clust generated {6} clusters of {1}s.' \
                 ''.format(label0, label1, L, dataset_datasets, Ng_original, Ng_to_clustering, K)
    res += msgformated(tmptxt)
    if Map is not None:
        tmptxt = '\n* An OG is a group of orthologous genes within & across different species, as identified by the ' \
                 'provided map.'
        res += msgformated(tmptxt)

    res += midline()

    res += msgformated('Citation\n~~~~~~~~', alignment='^')
    tmptxt = 'When publishing work that uses Clust, please include this citation:\n' \
             'Basel Abu-Jamous and Steven Kelly (2018) Clust: automatic extraction of optimal co-expressed ' \
             'gene clusters from gene expression data. Genome Biology 19:172; ' \
             'doi: https://doi.org/10.1186/s13059-018-1536-8.'
#TODO: citation
    res += msgformated(tmptxt)
    res += midline()

    tmptxt = 'For enquiries contact:\n' \
             'Dr. Basel Abu-Jamous\n' \
             'baselabujamous@gmail.com'
    res += msgformated(tmptxt)
    res += bottomline(withnewline=False)

    return res, endtime, timeConsumedTxt


def summarise_results(X, Xprocessed, Map, GDMall, GDM, uncle_res, mn_res,
                      B_corrected, starttime, endtime, timeconsumedtxt):
    K = B_corrected.shape[1]
    if K > 0:
        if Map is None:
            if len(X) == 1:
                res = collec.OrderedDict(
                    [('Starting data and time', starttime.strftime('%A %d %B %Y (%H:%M:%S)')),
                     ('Ending date and time', endtime.strftime('%A %d %B %Y (%H:%M:%S)')),
                     ('Time consumed', timeconsumedtxt),
                     ('Number of datasets', len(X)),
                     ('Total number of input genes', GDMall.shape[0]),
                     ('Genes included in the analysis', GDM.shape[0]),
                     ('Genes filtered out from the analysis', GDMall.shape[0] - GDM.shape[0]),
                     ('Number of clusters', B_corrected.shape[1]),
                     ('Total number of genes in clusters', np.sum(np.any(B_corrected, axis=1))),
                     ('Genes not included in any cluster', B_corrected.shape[0] - np.sum(np.any(B_corrected, axis=1))),
                     ('Min cluster size', np.min(np.sum(B_corrected, axis=0))),
                     ('Max cluster size', np.max(np.sum(B_corrected, axis=0))),
                     ('Average cluster size', np.mean(np.sum(B_corrected, axis=0))),
                     ('Number of samples (including replicates)', np.shape(X[0])[1]),
                     ('Number of conditions', np.shape(Xprocessed[0])[1]),
                     ])
            else:
                res = collec.OrderedDict(
                    [('Starting data and time', starttime.strftime('%A %d %B %Y (%H:%M:%S)')),
                     ('Ending date and time', endtime.strftime('%A %d %B %Y (%H:%M:%S)')),
                     ('Time consumed', timeconsumedtxt),
                     ('Number of datasets', len(X)),
                     ('Total number of input genes', GDMall.shape[0]),
                     ('Genes included in the analysis', GDM.shape[0]),
                     ('Genes filtered out from the analysis', GDMall.shape[0] - GDM.shape[0]),
                     ('Number of clusters', B_corrected.shape[1]),
                     ('Total number of genes in clusters', np.sum(np.any(B_corrected, axis=1))),
                     ('Genes not included in any cluster', B_corrected.shape[0] - np.sum(np.any(B_corrected, axis=1))),
                     ('Min cluster size', np.min(np.sum(B_corrected, axis=0))),
                     ('Max cluster size', np.max(np.sum(B_corrected, axis=0))),
                     ('Average cluster size', np.mean(np.sum(B_corrected, axis=0))),
                     ('Total number of samples (including replicates)', np.sum([np.shape(x)[1] for x in X])),
                     ('Average number of samples (including replicates) per dataset', np.mean([np.shape(x)[1] for x in X])),
                     ('Total number of conditions', np.sum([np.shape(x)[1] for x in Xprocessed])),
                     ('Average number of conditions per dataset', np.mean([np.shape(x)[1] for x in Xprocessed])),
                     ('nokey0', '**********************************'),
                     ('Genes included in all datasets', np.sum(np.all(GDMall, axis=1))),
                     ('Genes missed from 1 dataset', np.sum(np.sum(GDMall == 0, axis=1) == 1)),
                     ('Genes missed from 2 datasets', np.sum(np.sum(GDMall == 0, axis=1) == 2)),
                     ('Genes missed from 3 datasets', np.sum(np.sum(GDMall == 0, axis=1) == 3)),
                     ('Genes missed from more than 3 datasets', np.sum(np.sum(GDMall == 0, axis=1) > 3))
                     ])
        else:
            genesOGs = np.array([[len(og_species) for og_species in og] for og in Map]) # Number of genes per OG per species
            res = collec.OrderedDict(
                [('Starting data and time', starttime.strftime('%A %d %B %Y (%H:%M:%S)')),
                 ('Ending date and time', endtime.strftime('%A %d %B %Y (%H:%M:%S)')),
                 ('Time consumed', timeconsumedtxt),
                 ('Number of datasets', len(X)),
                 ('Total number of input OGs', GDMall.shape[0]),
                 ('OGs included in the analysis', GDM.shape[0]),
                 ('OGs filtered out from the analysis', GDMall.shape[0] - GDM.shape[0]),
                 ('Number of clusters', B_corrected.shape[1]),
                 ('Total number of OGs in clusters', np.sum(np.any(B_corrected, axis=1))),
                 ('OGs not included in any cluster', B_corrected.shape[0] - np.sum(np.any(B_corrected, axis=1))),
                 ('Min cluster size', np.min(np.sum(B_corrected, axis=0))),
                 ('Max cluster size', np.max(np.sum(B_corrected, axis=0))),
                 ('Average cluster size', np.mean(np.sum(B_corrected, axis=0))),
                 ('Total number of samples (including replicates)', np.sum([np.shape(x)[1] for x in X])),
                 ('Average number of samples (including replicates) per dataset', np.mean([np.shape(x)[1] for x in X])),
                 ('Total number of conditions', np.sum([np.shape(x)[1] for x in Xprocessed])),
                 ('Average number of conditions per dataset', np.mean([np.shape(x)[1] for x in Xprocessed])),
                 ('nokey0', '**********************************'),
                 ('Number of species', Map.shape[1]),
                 ('Average number of genes in an OG per type', np.mean(genesOGs)),
                 ('Max number of genes in an OG per type', np.max(genesOGs)),
                 ('Number of OGs with at least a single gene in each type', np.sum(np.all(genesOGs >= 1, axis=1))),
                 ('Number of OGs with exactly a single gene in each type', np.sum(np.all(genesOGs == 1, axis=1))),
                 ('nokey1', '**********************************'),
                 ('OGs included in all datasets', np.sum(np.all(GDMall, axis=1))),
                 ('OGs missed from 1 dataset', np.sum(np.sum(GDMall == 0, axis=1) == 1)),
                 ('OGs missed from 2 datasets', np.sum(np.sum(GDMall == 0, axis=1) == 2)),
                 ('OGs missed from 3 datasets', np.sum(np.sum(GDMall == 0, axis=1) == 3)),
                 ('OGs missed from more than 3 datasets', np.sum(np.sum(GDMall == 0, axis=1) > 3))
                 ])
    else:
        if Map is None:
            res = collec.OrderedDict(
                [('Starting data and time', starttime.strftime('%A %d %B %Y (%H:%M:%S)')),
                 ('Ending date and time', endtime.strftime('%A %d %B %Y (%H:%M:%S)')),
                 ('Time consumed', timeconsumedtxt),
                 ('Number of datasets', len(X)),
                 ('Total number of input genes', GDMall.shape[0]),
                 ('Genes included in the analysis', GDM.shape[0]),
                 ('Genes filtered out from the analysis', GDMall.shape[0] - GDM.shape[0]),
                 ('Number of clusters', B_corrected.shape[1]),
                 ('nokey0', '**********************************'),
                 ('Genes included in all datasets', np.sum(np.all(GDMall, axis=1))),
                 ('Genes missed from 1 dataset', np.sum(np.sum(GDMall == 0, axis=1) == 1)),
                 ('Genes missed from 2 datasets', np.sum(np.sum(GDMall == 0, axis=1) == 2)),
                 ('Genes missed from 3 datasets', np.sum(np.sum(GDMall == 0, axis=1) == 3)),
                 ('Genes missed from more than 3 datasets', np.sum(np.sum(GDMall == 0, axis=1) > 3))
                 ])
        else:
            genesOGs = np.array([[len(og_species) for og_species in og] for og in Map]) # Number of genes per OG per species
            meanGenesOGs = 0 if len(genesOGs) == 0 else np.mean(genesOGs)
            maxGenesOGs = 0 if len(genesOGs) == 0 else np.max(genesOGs)
            OGsWithAtLeastASingleGeneInEachType = 0 if len(genesOGs) == 0 else np.sum(np.all(genesOGs >= 1, axis=1))
            OGsWithExactlyASingleGeneInEachType = 0 if len(genesOGs) == 0 else np.sum(np.all(genesOGs == 1, axis=1))
            res = collec.OrderedDict(
                [('Starting data and time', starttime.strftime('%A %d %B %Y (%H:%M:%S)')),
                 ('Ending date and time', endtime.strftime('%A %d %B %Y (%H:%M:%S)')),
                 ('Time consumed', timeconsumedtxt),
                 ('Number of datasets', len(X)),
                 ('Total number of input OGs', GDMall.shape[0]),
                 ('OGs included in the analysis', GDM.shape[0]),
                 ('OGs filtered out from the analysis', GDMall.shape[0] - GDM.shape[0]),
                 ('Number of clusters', B_corrected.shape[1]),
                 ('nokey0', '**********************************'),
                 ('Number of species', Map.shape[1]),
                 ('Average number of genes in an OG per type', meanGenesOGs),
                 ('Max number of genes in an OG per type', maxGenesOGs),
                 ('Number of OGs with at least a single gene in each type', OGsWithAtLeastASingleGeneInEachType),
                 ('Number of OGs with exactly a single gene in each type', OGsWithExactlyASingleGeneInEachType),
                 ('nokey1', '**********************************'),
                 ('OGs included in all datasets', np.sum(np.all(GDMall, axis=1))),
                 ('OGs missed from 1 dataset', np.sum(np.sum(GDMall == 0, axis=1) == 1)),
                 ('OGs missed from 2 datasets', np.sum(np.sum(GDMall == 0, axis=1) == 2)),
                 ('OGs missed from 3 datasets', np.sum(np.sum(GDMall == 0, axis=1) == 3)),
                 ('OGs missed from more than 3 datasets', np.sum(np.sum(GDMall == 0, axis=1) > 3))
                 ])
    return res


def clusters_genes_OGs(B, OGs, Map, MapSpecies, delim='; '):
    if Map is None:
        Nsp = 0
    else:
        Nsp = len(MapSpecies)  # Number of species
    K = B.shape[1]  # Number of clusters
    if K == 0:
        return pd.DataFrame(data=np.array(np.empty([0, 0]), dtype=object), index=None, columns=None, dtype=str)
    Csizes = np.sum(B, axis=0)  # Clusters' sizes
    maxCsize = np.max(Csizes)  # Largest cluster size
    res = np.array(np.empty([maxCsize, (Nsp + 1) * K], dtype=str), dtype=object)
    header = np.array([None] * ((Nsp + 1) * K * 2), dtype=object).reshape([2, ((Nsp + 1) * K)])
    for k in range(K):
        col = k * (Nsp + 1)
        res[0:Csizes[k], col] = OGs[B[:, k]]
        res[Csizes[k]:, col] = ''
        header[0, col] = 'C{0} ({1} {2})'.format(k, Csizes[k], 'genes' if Map is None else 'OGs')
        header[1, col] = 'Genes' if Map is None else 'OGs'
        for sp in range(Nsp):  # Will not get into this if Map is None, as Nsp will be zero in that case
            col = k * (Nsp + 1) + sp + 1
            res[0:Csizes[k], col] = [ds.concatenateStrings(gs, delim) for gs in Map[B[:, k], sp]]
            res[Csizes[k]:, col] = ''
            header[0, col] = ''
            header[1, col] = MapSpecies[sp]


    return pd.DataFrame(data=np.array(np.concatenate((header, res), axis=0), dtype=str), index=None, columns=None, dtype=str)


def clusters_genes_Species(B, OGs, Map, MapSpecies):
    Nsp = len(MapSpecies)  # Number of species
    K = B.shape[1]  # Number of clusters

    # If no clusters, return a trivial output
    if K == 0:
        resFrames = np.array([None] * Nsp, dtype=object)
        for sp in range(Nsp):
            resFrames[sp] = pd.DataFrame(data=np.empty([0, 0]), columns=None, index=None, dtype=str)
        return resFrames

    # Find flattened genes in species
    flatGenesInSpecies = [[ds.flattenAList(Map[B[:, k], sp]) for k in range(K)] for sp in range(Nsp)]  # Nsp x K lists

    # Prepare the results object
    Csizes = [[len(sp_k_genes) for sp_k_genes in sp_genes] for sp_genes in flatGenesInSpecies]  # Nsp x K
    maxCsizes = [np.max(csizes_sp) for csizes_sp in Csizes]  # Nsp x 1
    res = np.array([None] * Nsp, dtype=object)
    resFrames = np.array([None] * Nsp, dtype=object)

    # Fill the results object, species by species
    for sp in range(Nsp):
        restmp = np.array(np.empty([maxCsizes[sp], K], dtype=str), dtype=object)
        header = np.array([None] * K, dtype=object).reshape([1, K])
        for k in range(K):
            restmp[0:Csizes[sp][k], k] = flatGenesInSpecies[sp][k]
            restmp[Csizes[sp][k]:, k] = ''
            header[0, k] = 'C{0} ({1} genes)'.format(k, Csizes[sp][k])
        res[sp] = np.array(np.concatenate((header, restmp), axis=0), dtype=str)
        resFrames[sp] = pd.DataFrame(data=res[sp], columns=None, index=None, dtype=str)

    return resFrames


'''
B: Ng x K
OGs: Ng x 1
Map: Ng x Nspecies
MapSpecies: array of names of species in the Map (header of Map)
'''
def clusters_B_as_dataframes(B, OGs, Map):
    clusterslabels = ['C{0}'.format(c) for c in range(np.shape(B)[1])]
    B = pd.DataFrame(B, OGs, clusterslabels, dtype=bool)

    if Map is None:
        return B
    else:
        Nsp = np.shape(Map)[1]
        PerSpecies_B = [None] * Nsp
        for sp in range(Nsp):
            RowNames = [ds.concatenateStrings(gs, ',') for gs in Map[:, sp]]  # Groups of genes
            PerSpecies_B[sp] = pd.DataFrame(B, RowNames, clusterslabels, dtype=bool)

        return B, PerSpecies_B


'''
B: Ng x K
OGs: Ng x 1
Map: Ng x Nspecies
MapSpecies: array of names of species in the Map (header of Map)
'''
def processed_X_as_dataframes(X_processed, OGs, conditions):
    L = len(X_processed)

    Xout = [None] * L

    for l in range(L):
        Xout[l] = pd.DataFrame(X_processed[l], OGs, conditions[l])

    return Xout



def processed_X(Xprocessed, conditions, GDM, OGs, Map, MapSpecies):
    L = len(Xprocessed)
    res = np.array([None] * L, dtype=object)
    resFrame = np.array([None] * L, dtype=object)
    resData = np.array([None] * L, dtype=object)
    resHeader = np.array([None] * L, dtype=object)
    resGeneNames = np.array([None] * L, dtype=object)

    for l in range(L):
        # Header (Samples)
        #samploc = np.array(Samples[l])
        #uniqueSamploc = np.unique(SamplesIDs[l])
        #uniqueSamploc = uniqueSamploc[uniqueSamploc >= 0]
        #resHeader[l] = [samploc[np.array(SamplesIDs[l]) == s][0] for s in uniqueSamploc]
        resHeader[l] = conditions[l]
        if Map is None:
            resHeader[l] = np.array([['Genes'] + resHeader[l]])
        else:
            resHeader[l] = np.array([['OGs'] + MapSpecies.tolist() + resHeader[l]])

        # Gene names
        if Map is None:
            resGeneNames[l] = OGs[GDM[:, l]].reshape(-1,1)
        else:
            genenames = [[ds.concatenateStrings(gs) for gs in Map[GDM[:, l]][:, sp]] for sp in range(Map.shape[1])]
            resGeneNames[l] = np.concatenate((OGs[GDM[:, l]].reshape(-1,1), np.transpose(genenames)), axis=1)

        # Data
        resData[l] = np.array(Xprocessed[l])

        # concatenate them
        res[l] = np.concatenate((resGeneNames[l], resData[l]), axis=1)
        res[l] = np.concatenate((resHeader[l], res[l]), axis=0)
        res[l] = np.array(res[l], dtype=str)

        resFrame[l] = pd.DataFrame(data=res[l], index=None, columns=None, dtype=str)

    return resFrame


def params(params, q3s, OGsIncludedIfAtLeastInDatasets, expressionValueThreshold,
           atleastinconditions, atleastindatasets, deterministic, methods, Map=None):
    if Map is None:
        res = collec.OrderedDict(
            [('Methods', params['methods']),
             ('K values (Ks)', params['Ks']),
             ('Tightness weight', params['tightnessweight']),
             ('Q3s outliers threshold', q3s),
             ('Genes included if at least in datasets', OGsIncludedIfAtLeastInDatasets),
             ('Filtering: Data value threshold', expressionValueThreshold),
             ('Filtering: At least in conditions', atleastinconditions),
             ('Filtering: At least in datasets', atleastindatasets),
             ('Is deterministic?', deterministic),
             ('Seed methods', methods)
             ])
    else:
        res = collec.OrderedDict(
            [('Methods', params['methods']),
             ('K values (Ks)', params['Ks']),
             ('Tightness weight', params['tightnessweight']),
             ('Q3s outliers threshold', q3s),
             ('OGs included if at least in datasets', OGsIncludedIfAtLeastInDatasets),
             ('Filtering: Data value threshold', expressionValueThreshold),
             ('Filtering: At least in conditions', atleastinconditions),
             ('Filtering: At least in datasets', atleastindatasets),
             ('Is deterministic?', deterministic),
             ('Seed methods', methods)
             ])
    return res



# HOW TO SETUP (FOR PIP)
# Go to the clust directory and remove dist/ and build folders by
# sudo rm -r dist/
# sudo rm -r build/
# Create the bdist_wheel file:
# sudo python3 setup.py sdist bdist_wheel
# Upload to pypi:
# twine upload dist/*
# You will be asked for username (basel) and pword
#
# Updating bioconda:
# Edit this GitHub file:
# https://github.com/bioconda/bioconda-recipes/edit/master/recipes/clust/meta.yaml
# Namely edit the version number and the sha256 number.
#
# Get the sha256 number by this in the command line:
# wget -O- $URL | shasum -a 256
# Where the URL is this (with {{ version }} filled:
# https://pypi.io/packages/source/c/clust/clust-{{ version }}.tar.gz
#
# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
import sys
from os import path
#from clust.scripts.glob import version
version='v1.13.0'


def setupmain(args=None):
    if args is None:
        args = sys.argv[1:]

    here = path.abspath(path.dirname(__file__))

    setup(
        name='clust',
        version=version,

        description='Optimised consensus clustering of multiple heterogeneous datasets',
        long_description='Optimised consensus clustering of multiple heterogeneous datasets',

        # The project's main homepage.
        url='https://github.com/baselabujamous/clust',

        # Author details
        author='Basel Abu-Jamous',
        author_email='baselabujamous@gmail.com',

        # Choose your license
        license='OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',

        classifiers=[
            # How mature is this project? Common values are
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            'Development Status :: 3 - Alpha',

            # Indicate who your project is intended for
            'Intended Audience :: Science/Research',
            'Topic :: Software Development :: Build Tools',

            # Pick your license as you wish (should match "license" above)
            'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',

            # Specify the Python versions you support here.
            'Programming Language :: Python :: 3',
        ],

        # What does your project relate to?
        keywords='',

        # You can just specify the packages manually here if your project is
        # simple. Or you can use find_packages().
        packages=find_packages(exclude=['docs', 'tests']),

        # Alternatively, if you want to distribute just a my_module.py, uncomment
        # this:
        #   py_modules=["my_module"],

        # List run-time dependencies here.  These will be installed by pip when
        # your project is installed. For an analysis of "install_requires" vs pip's
        # requirements files see:
        # https://packaging.python.org/en/latest/requirements.html
        install_requires=[
            'joblib', 
            'portalocker'
            'numpy', 
            'scipy', 
            'pandas', 
            'matplotlib', 
            'scikit-learn',
        ],

        # If there are data files included in your packages that need to be
        # installed, specify them here.  If using Python 2.6 or less, then these
        # have to be included in MANIFEST.in as well.
        # package_data={
        #     'sample': ['package_data.dat'],
        # },

        # Although 'package_data' is the preferred approach, in some case you may
        # need to place data files outside of your packages. See:
        # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
        # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
        # data_files=[('my_data', ['data/data_file'])],

        # To provide executable scripts, use entry points in preference to the
        # "scripts" keyword. Entry points provide cross-platform support and allow
        # pip to create the appropriate form of executable for the target platform.
        entry_points={
            'console_scripts': [
                'clust=clust.__main__:main',
            ],
        },
    )


if __name__ == "__main__":
    setupmain()




import os
import re
import numpy as np
import clust.scripts.datastructures as ds
import clust.scripts.numeric as nu
import clust.scripts.glob as glob
import clust.scripts.output as op
import sys
import traceback
import math
import portalocker
import pandas as pd


if sys.version_info[0] < 3:
    from StringIO import StringIO
else:
    from io import StringIO


def getFilesInDirectory(path, extension=None):
    for (dirpath, dirnames, filenames) in os.walk(path):
        if extension is None or extension == '':
            return [fn for fn in filenames if fn != '.DS_Store']
        else:
            if len(extension) > 1 and extension[0] == '.' and extension[1] != '*':
                extension = extension[1:]
            return [fn for fn in filenames if re.match('(.*\.' + extension + '$)', fn) is not None and fn != '.DS_Store']


def readDatasetsFromDirectory(path, delimiter='\t| |, |; |,|;', skiprows=1, skipcolumns=1, returnSkipped=False):

    # Single file name given
    if os.path.isfile(path):
        datafiles = [os.path.basename(path)]
        datafileswithpath = [path]
    # Directory of data files given
    elif os.path.isdir(path):
        datafiles = np.sort(getFilesInDirectory(path)).tolist()
        datafileswithpath = [path + '/' + df for df in datafiles]
    # Invalid path given
    else:
        raise ValueError('Data path {0} does not exist. Either provide a path '.format(path) + \
                         'of a data file or a path to a directory including data file(s)')

    datafilesread = readDataFromFiles(datafileswithpath, delimiter, float, skiprows, skipcolumns, returnSkipped)

    if returnSkipped:
        return datafilesread + (datafiles, )
    else:
        return datafilesread, datafiles


def readMap(mapfile, delimiter='\t'):
    if mapfile is None:
        return None
    return readDataFromFiles([mapfile], delimiter, dtype=str, skiprows=0, skipcolumns=0, returnSkipped=False, data_na_filter=False)[0]


def readReplicates(replicatesfile, datapath, datafiles, replicates, delimiter='\t| |,|;'):
    if replicatesfile is None:
        if isinstance(replicates[0], list):
            return None, replicates
        elif isinstance(replicates[0], np.ndarray):
            return None, [r.tolist() for r in replicates]
        else:
            return None, replicates

    L = len(datafiles)
    replicatesIDs = [[-1 for c in replicates[l]] for l in range(L)]
    conditions = [None] * L

    with open(replicatesfile) as f:
        lineNumber = 0
        for line in f:
            lineNumber += 1
            line = line.partition('#')[0]
            line = line.rstrip()
            line = list(filter(None, re.split(delimiter, line)))

            # Skip to next line if it is an empty line
            if len(line) < 1:
                continue

            if line[0] in datafiles:
                l = datafiles.index(line[0])  # (l)th dataset
            else:
                raise ValueError('Unrecognised data file name ({0}) in line {1} in {2}.'.
                                 format(line[0], lineNumber, replicatesfile))

            # Skip to next line if no condition ID is given
            if len(line) < 2:
                continue

            if conditions[l] is None:
                conditions[l] = [line[1]]
            elif line[1] not in conditions:
                conditions[l] += [line[1]]

            c = conditions[l].index(line[1])  # (c)th condition

            # Skip to next line if no replicates are given
            if len(line) < 3:
                continue

            for r in line[2:]:
                if r in replicates[l]:
                    if isinstance(replicates[l], list):
                        replicatesIDs[l][replicates[l].index(r)] = c
                    elif isinstance(replicates[l], np.ndarray):
                        replicatesIDs[l][replicates[l].tolist().index(r)] = c
                else:
                    raise ValueError('Unrecognised replicate name ({0}) in line {1} in {2}.'.
                                     format(r, lineNumber, replicatesfile))

    # Write the conditions of the data files with no entries in the replicates file
    for c in range(len(conditions)):
        if conditions[c] is None:
            with open('{0}/{1}'.format(datapath, datafiles[c])) as f:
                line = f.readline()
                while (line[0] == '#'):
                    line = f.readline()
                line = line.rstrip()
                line = filter(None, re.split(delimiter, line))
                conditions[c] = line[1:]
            replicatesIDs[c] = list(range(len(conditions[c])))

    return replicatesIDs, conditions


def readNormalisation(normalisefile, datafiles, delimiter='\t| |,|;', defaultnormalisation=1000):
    """

    :param normalisefile: either a list of a single string element which is the normalisation file name,
    or a list of strings representing normalisation codes. In this case, the strings must be convertable to integers.
    :param datafiles:
    :param delimiter:
    :param defaultnormalisation:
    :return:
    """
    if normalisefile is None:
        return defaultnormalisation

    # This is in case the normalisation file was given as a single integer, it should not though
    if nu.isint(normalisefile):
        normalisefile = [normalisefile]

    L = len(datafiles)
    normalise = [None] * L

    # This happens when the normalisation codes are given directly rather than in a file
    if len(normalisefile) > 1 or nu.isint(normalisefile[0]):
        for l in range(L):
            normalise[l] = [int(n) for n in normalisefile]
        return normalise

    # This happens when a normalisation file is given
    with open(normalisefile[0]) as f:
        lineNumber = 0
        for line in f:
            lineNumber += 1
            line = line.partition('#')[0]
            line = line.rstrip()
            line = list(filter(None, re.split(delimiter, line)))

            # Skip to next line if it is an empty line
            if len(line) < 1:
                continue

            if line[0] in datafiles:
                l = datafiles.index(line[0])  # (l)th dataset
            else:
                raise ValueError('Unrecognised data file name ({0}) in line {1} in {2}.'.
                                 format(line[0], lineNumber, normalisefile[0]))

            # If no normalisation is set for the dataset, skip to the next line
            if len(line) < 2:
                continue

            # If the normalisation of this dataset has not been set, set it, otherwise append
            if normalise[l] is None:
                normalise[l] = line[1:]
            else:
                normalise[l] = normalise[l] + line[1:]

    for l in range(L):
        if normalise[l] is None:
            normalise[l] = [defaultnormalisation]
        else:
            normalise[l] = [int(n) for n in normalise[l]]

    return normalise


def readDataFromFiles(datafiles, delimiter='\t| |, |; |,|;', dtype=float, skiprows=1, data_na_filter=True, skipcolumns=1, returnSkipped=True, comm='#'):
    L = len(datafiles)
    X = [None] * L
    skippedRows = [None] * L
    skippedCols = [None] * L
    for l in range(L):
        with open(datafiles[l]) as f:
            ncols = len(re.split(delimiter, f.readline()))
        # This is now using pandas read_csv, if np.loadtxt is re-used, you HAVE TO set ndmin = 2 here
        X[l] = pdreadcsv_regexdelim(datafiles[l], delimiter=delimiter, dtype=dtype, skiprows=skiprows,
                          usecols=range(skipcolumns, ncols), na_filter=data_na_filter, comments=comm)

        if skiprows > 0:
            skippedRows[l] = pdreadcsv_regexdelim(datafiles[l], delimiter=delimiter, dtype=str, skiprows=0,
                                        usecols=range(skipcolumns, ncols), na_filter=False, comments=comm)[0:skiprows]
            if skiprows == 1:
                skippedRows[l] = skippedRows[l][0]
        else:
            skippedRows[l] = np.array([]).reshape([0, X[l].shape[1]])

        if skipcolumns > 0:
            skippedCols[l] = pdreadcsv_regexdelim(datafiles[l], delimiter=delimiter, dtype=str, skiprows=skiprows,
                                        usecols=range(skipcolumns), na_filter=False, comments=comm)
        else:
            skippedCols[l] = np.array([]).reshape([0, X[l].shape[1]])

    if returnSkipped:
        return (ds.listofarrays2arrayofarrays(X), skippedRows, skippedCols)
    else:
        return ds.listofarrays2arrayofarrays(X)


# Does the same job as numpy.loadtxt but regex delimiter
#### OBSOLETE, USE THE PANDAS READ_CSV VERSION BELOW INSTEAD
def nploadtxt_regexdelim(file, delimiter='\t| |, |; |,|;', dtype=float, skiprows=0, usecols=None, ndmin=0, comments='#'):
    with open(file) as f:
        result = np.loadtxt((re.sub(delimiter, '\t', str(x)) for x in f),
                            delimiter='\t', dtype=dtype, skiprows=skiprows, usecols=usecols, ndmin=ndmin, comments=comments)
    return result


# Does the same job as pandas.read_csv but regex delimiter
def pdreadcsv_regexdelim(file, delimiter='\t| |, |; |,|;', dtype=float, skiprows=0, usecols=None, na_filter=True, comments='#'):
    with open(file) as f:
        result = pd.read_csv(StringIO('\n'.join(re.sub(delimiter, '\t', str(x)) for x in f)),
                            delimiter='\t', dtype=dtype, header=None, skiprows=skiprows, usecols=usecols, na_filter=na_filter, comment=comments).values
    return result


def writedic(filepath, dic, header=None, delim='\t'):
    f = open(filepath, 'w+')

    # Write header
    if header is not None:
        f.write('{0}\n'.format(header))

    # Write the rest
    nokey = re.compile('^nokey[0-9]*$', flags=re.I)  # To match lines with no keys
    for k in dic.keys():

        if nokey.match(k) is None:
            f.write('{0}{1}{2}\n'.format(k, delim, dic[k]))
        else:
            f.write('{0}\n'.format(dic[k]))

    # Close file
    f.close()


def log(msg=None, addextrastick=True):
    if addextrastick:
        msg = op.msgformated(msg, withnewline=False)

    if glob.print_to_log_file:
        with open(glob.logfile, mode='a+') as f:
            if msg is not None:
                f.write(msg)
            f.write('\n')

    if glob.print_to_console:
        print(msg)


def logerror(exec_info):
    errstr = traceback.format_exception(exec_info[0], exec_info[1], exec_info[2])
    errstr = ''.join(errstr)
    log('Unexpected error:\n{0}\nContinuing execution anyway ...'.format(errstr))


def resetparallelprogress(parallel_total, log_every_percentage=10.0):
    with open(glob.tmpfile, mode='w+') as f:
        f.write('{0} {1} {2}'.format(parallel_total, log_every_percentage, 0.0))
        f.truncate()


def updateparallelprogress(added_value):
    Done = False
    while (not Done):
        try:
            Done = True
            with open(glob.tmpfile, mode='r+') as f:
                portalocker.lock(f, portalocker.LOCK_EX)
                data = f.read().split(" ")
                parallel_total = float(data[0])
                log_every_percentage = float(data[1])
                current_parallel_progress = float(data[2])

                last_log = math.floor(100 * current_parallel_progress
                                      / parallel_total / log_every_percentage) * log_every_percentage
                current_parallel_progress += added_value
                new_log = math.floor(100 * current_parallel_progress
                                     / parallel_total / log_every_percentage) * log_every_percentage

                #for i in np.arange(last_log+log_every_percentage, new_log + log_every_percentage, log_every_percentage):
                #    log('{0}%'.format(int(i)))
                if new_log > last_log:
                    log('{0}%'.format(int(new_log)))

                f.seek(0)
                f.write('{0} {1} {2}'.format(parallel_total, log_every_percentage, current_parallel_progress))
                f.truncate()
        except:
            Done = False


def getparallelprogress():
    Done = False
    while (not Done):
        try:
            Done = True
            with open(glob.tmpfile, mode='r+') as f:
                data = f.read().split(" ")
                parallel_total = float(data[0])
                log_every_percentage = float(data[1])
                current_parallel_progress = float(data[2])

                return (parallel_total, log_every_percentage, current_parallel_progress)
        except:
            Done = False


def deletetmpfile():
    if os.path.isfile(glob.tmpfile):
        os.remove(glob.tmpfile)




import functools
import math
import numpy as np
import sklearn.metrics.pairwise as skdists


def factors_primesunique(n):
    return [i for i in primes(math.ceil(math.sqrt(n))) if n % i == 0]


def factors_primes(n):
    res = factors_primesunique(n)
    n /= np.prod(res)
    while n > 1:
        localres = factors_primesunique(n)
        res += localres
        n /= np.prod(localres)
    return np.sort(res).tolist()


def factors_all(n):
    return np.unique(functools.reduce(list.__add__,
                                         ([i, n // i] for i in range(1, int(n ** 0.5) + 1) if n % i == 0))).tolist()


def is_prime(n):
    for j in range(2, math.ceil(math.sqrt(n)) + 1):
        if (n % j) == 0:
            return False
    return True


def primes(n):
    if n <= 1:
        return []
    elif n == 2 or n == 3:
        return list(range(2, n + 1))
    else:
        res = [2, 3]
        for i in range(5, n + 1, 2):
            if is_prime(i):
                res.append(i)
    return res


def closest_to_square_factors(n):
    f = factors_all(n)
    sqrtn = math.sqrt(n)
    min_ind = np.argmin([abs(x-sqrtn) for x in f])
    res = [f[min_ind], int(n/f[min_ind])]
    if res[0] > res[1]:
        tmp = res[0]
        res[0] = res[1]
        res[1] = tmp
    return res


def isint(s):
    """
    Test if string s is an integer
    :param s: string
    :return: boolean
    """
    try:
        int(s)
        return True
    except ValueError:
        return False
    except TypeError:
        return False


def getcondvects(n):
    g = 1
    n2 = 2 ** n
    condvects = np.zeros((n2, n), dtype=bool)
    for m in range(1,n+1):
        m2 = 2 ** m
        m3 = int(m2 / 2)
        n3 = n - m
        for g in range(g, n2, m2):
            for k in range(m3):
                condvects[g + k, n3] = 1
        g = m2
    return condvects


def addaxis(X, V, axis=0):
    Xloc = np.array(X)
    if axis == 0:
        return np.add(Xloc, V)
    if axis == 1:
        return np.add(Xloc.transpose(), V).transpose()
    raise ValueError('Invalid axis value; it has to be 0 or 1')


def subtractaxis(X, V, axis=0):
    Xloc = np.array(X)
    if axis == 0:
        return np.subtract(Xloc, V)
    if axis == 1:
        return np.subtract(Xloc.transpose(), V).transpose()
    raise ValueError('Invalid axis value; it has to be 0 or 1')


def multiplyaxis(X, V, axis=0):
    Xloc = np.array(X)
    if axis == 0:
        return np.multiply(Xloc, V)
    if axis == 1:
        return np.multiply(Xloc.transpose(), V).transpose()
    raise ValueError('Invalid axis value; it has to be 0 or 1')


def divideaxis(X, V, axis=0):
    Xloc = np.array(X)
    np.seterr(divide='ignore', invalid='ignore')
    if axis == 0:
        return np.divide(Xloc, V)
    if axis == 1:
        return np.divide(Xloc.transpose(), V).transpose()
    raise ValueError('Invalid axis value; it has to be 0 or 1')


def isequaltoaxis(X, V, axis=0):
    Xloc = np.array(X)
    [N,M] = Xloc.shape
    if axis == 0:
        if len(V) != M:
            raise ValueError('The length of the vector V should be equal to the number of columns in the matrix X')
        Vloc = np.tile(V, [N, 1])
        return Xloc == Vloc
    if axis == 1:
        if len(V) != N:
            raise ValueError('The length of the vector V should be equal to the number of rows in the matrix X')
        Vloc = np.tile(V, [M, 1]).transpose()
        return Xloc == Vloc
    raise ValueError('Invalid axis value; it has to be 0 or 1')


def lessthanaxis(X, V, axis=0, orequal=False):
    Xloc = np.array(X)
    [N,M] = Xloc.shape
    if axis == 0:
        if len(V) != M:
            raise ValueError('The length of the vector V should be equal to the number of columns in the matrix X')
        Vloc = np.tile(V, [N, 1])
        if (orequal):
            return Xloc <= Vloc
        else:
            return Xloc < Vloc
    if axis == 1:
        if len(V) != N:
            raise ValueError('The length of the vector V should be equal to the number of rows in the matrix X')
        Vloc = np.tile(V, [M, 1]).transpose()
        if (orequal):
            return Xloc <= Vloc
        else:
            return Xloc < Vloc
    raise ValueError('Invalid axis value; it has to be 0 or 1')


def largerthanaxis(X, V, axis=0, orequal=False):
    Xloc = np.array(X)
    [N,M] = Xloc.shape
    if axis == 0:
        if len(V) != M:
            raise ValueError('The length of the vector V should be equal to the number of columns in the matrix X')
        Vloc = np.tile(V, [N, 1])
        if (orequal):
            return Xloc >= Vloc
        else:
            return Xloc > Vloc
    if axis == 1:
        if len(V) != N:
            raise ValueError('The length of the vector V should be equal to the number of rows in the matrix X')
        Vloc = np.tile(V, [M, 1]).transpose()
        if (orequal):
            return Xloc >= Vloc
        else:
            return Xloc > Vloc
    raise ValueError('Invalid axis value; it has to be 0 or 1')


def all_perms(elements):
    def local_all_perms(elements):
        if len(elements) <=1:
            yield elements
        else:
            for perm in all_perms(elements[1:]):
                for i in range(len(elements)):
                    # nb elements[0:1] works in both string and list contexts
                    yield perm[:i] + elements[0:1] + perm[i:]
    return list(local_all_perms(elements))


def dist_matrices(X1, X2, criterion='euclidean'):
    X1loc = np.array(X1)
    X2loc = np.array(X2)

    if len(X1loc.shape) == 1:
        if len(X2loc.shape) == 1:
            if X1loc.shape[0] == X2loc.shape[0]:
                # As row vectors
                X1loc = X1loc.reshape(1, -1)
                X2loc = X2loc.reshape(1, -1)
            else:
                # As column vectors
                X1loc = X1loc.reshape(-1, 1)
                X2loc = X2loc.reshape(-1, 1)
        else:
            if X1loc.shape[0] == X2loc.shape[1]:
                # Row vector VS. Many rows
                X1loc = X1loc.reshape(1, -1)
            elif X2loc.shape[1] == 1:
                # Column vector VS. Column vector
                X1loc = X1loc.reshape(-1, 1)
            elif X1loc.shape[0] == X2loc.shape[0]:
                # Row vector VS. transposed columns
                X1loc = X1loc.reshape(1, -1)
                X2loc = X2loc.transpose()
            else:
                raise ValueError('Invalid dimensions of X1 and X2')
    elif len(X2loc.shape) == 1:
        if X2loc.shape[0] == X1loc.shape[1]:
            # Many rows VS. row vector
            X2loc = X2loc.reshape(1, -1)
        else:
            raise ValueError('Invalid dimensions of X1 and X2')

    if criterion == 'euclidean':
        return skdists.euclidean_distances(X1loc, X2loc)
    elif criterion == 'hamming':
        raise NotImplementedError('Hamming distance between rows of matrices has not been implemented yet.')
    else:
        raise ValueError('Invalid distance criterion')